import warnings


def cache_page(*args, **kwargs):
    warnings.warn(
        'cache_page is deprecated and no longer needed to use cache, '
        'it will be removed in rmr-django 2.0',
        DeprecationWarning,
    )

    def _decorator(fn):
        return fn
    return _decorator



import warnings


class UpdateCacheMiddleware:

    def __init__(self):
        warnings.warn(
            'UpdateCacheMiddleware no longer needed to use cache '
            'and will be removed in rmr-django 2.0',
            RuntimeWarning,
        )


class CacheMiddleware:

    def __init__(self):
        warnings.warn(
            'CacheMiddleware no longer needed to use cache '
            'and will be removed in rmr-django 2.0',
            RuntimeWarning,
        )



import os


class VersionMiddleware:

    @staticmethod
    def process_response(request, response):
        version = os.environ.get('APP_VERSION')
        if version:
            response['App-Version'] = version
        return response



import warnings

from rmr.middleware.version import VersionMiddleware

warnings.warn(
    'VersionMiddleware moved to rmr.middleware.version, make sure '
    'you import it from the new place before upgrading to rmr-django 2.0',
    category=RuntimeWarning,
)



import warnings

from django import forms


class MultipleValueField(forms.TypedMultipleChoiceField):

    def valid_value(self, value):
        return True

    def to_python(self, value):
        return value


class MultiValueField(MultipleValueField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "MultiValueField is deprecated and will be removed "
            "in rmr-django 2.0use MultipleValueField instead",
            DeprecationWarning,
        )
        super(MultipleValueField, self).__init__(*args, **kwargs)


class BooleanField(forms.Field):

    empty_values = {None}

    def to_python(self, value):
        if value in (True, False):
            return value
        return None



import functools
import warnings


def replace_if(condition, replacement):
    """
    If `condition(...)` is True then returns result of `replacement(...)`
    evaluation and result of original function otherwise
    """
    def _decorator(fn):
        @functools.wraps(fn)
        def _fn(*args, **kwargs):
            if condition(*args, **kwargs):
                return replacement(*args, **kwargs)
            return fn(*args, **kwargs)
        return _fn
    return _decorator


def conditional(condition, fn_true=None, fn_false=None):
    """
    Depending on result of `condition` executes either `fn_true` or `fn_false`
    """
    warnings.warn(
        '`conditional` decorator is deprecated and will be removed '
        'in rmr-django 2.0, use `replace_if` instead',
        DeprecationWarning,
    )

    def _decorator(fn):
        @functools.wraps(fn)
        def _fn(*args, **kwargs):
            callback = fn
            if condition(*args, **kwargs):
                if callable(fn_true):
                    callback = fn_true
            elif callable(fn_false):
                callback = fn_false
            return callback(*args, **kwargs)
        return _fn
    return _decorator



import warnings

from django import forms


class OffsetLimit(forms.Form):

    offset = forms.IntegerField(required=False, min_value=0)
    limit = forms.IntegerField(required=False, min_value=1)

    @classmethod
    def with_limit_required(cls, *, max_value=None, **kwargs):
        if 'limit_max_value' in kwargs:
            warnings.warn(
                'limit_max_value deprecated and will be removed '
                'in rmr-django 2.0, use max_value instead',
                category=RuntimeWarning, stacklevel=2,
            )
            max_value = kwargs['limit_max_value']
        limit = forms.IntegerField(
            required=True,
            max_value=max_value,
            min_value=1,
        )
        return type('OffsetLimitRequired', (cls, ), dict(limit=limit))



import warnings

from collections import defaultdict


def dictfetchall(cursor):
    """Return all rows from a cursor as a dict"""
    columns = [col[0] for col in cursor.description]
    return (
        dict(zip(columns, row))
        for row in cursor.fetchall()
    )


class BulkModelCreator:
    """
    BulkModelCreator context manager accumulates instances of any Django model
    and flushes them to the DB (using bulk_create) when count
    of particular model instances reaches 'batch_size' parameter.
    All remaining items will be proceeded after exiting context.
    """

    def __init__(self, batch_size=300):
        self.batches = defaultdict(list)
        self.batch_size = batch_size

    def add(self, instance):
        warnings.warn(
            'BulkModelCreator.add is deprecated and will be removed '
            'in rmr-django 2.0, use BulkModelCreator.append instead',
            DeprecationWarning,
        )
        flushed = False
        self.batches[instance.__class__].append(instance)
        if len(self.batches[instance.__class__]) == self.batch_size:
            self.flush(instance.__class__)
            flushed = True
        return flushed

    def append(self, instance):
        self.batches[instance.__class__].append(instance)
        if len(self.batches[instance.__class__]) == self.batch_size:
            self.flush(instance.__class__)

    def flush(self, model_class):
        model_class.objects.bulk_create(
            self.batches[model_class],
            self.batch_size,
        )
        self.batches[model_class].clear()

    def flush_all(self):
        for key in self.batches:
            self.flush(key)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.flush_all()



import crcmod


def crc32(string, make_crc32=crcmod.predefined.mkPredefinedCrcFun('crc-32')):
    crc = make_crc32(bytes(string, encoding='utf-8'))
    if crc > 2147483647:  # 2 ** 32 // 2 - 1
        crc = -(crc >> 1)
    return crc


def crc64(string, make_crc64=crcmod.predefined.mkPredefinedCrcFun('crc-64')):
    crc = make_crc64(bytes(string, encoding='utf-8'))
    if crc > 9223372036854775807:  # 2 ** 64 // 2 - 1
        crc = -(crc >> 1)
    return crc



import collections
import itertools


def split_every(iterable, chunk_size):
    """
    Split iterable into chunks with given chunk_size
    """
    iterator = iter(iterable)
    while True:
        chunk = itertools.islice(iterator, chunk_size)
        first_item = next(chunk)  # raises StopIteration
        chunk = itertools.chain([first_item], chunk)
        yield chunk
        consume(chunk)  # ensure chunk is exhausted


def unique(iterable):
    return collections.OrderedDict.fromkeys(iterable).keys()


def multimap(fn_list, *iterables):
    """
    Sequentially applies all callable from `fn_list`
    to every item of `iterables`, yielding the results
    """
    for args_set in zip(*iterables):
        fn_list, fn_list_copy = itertools.tee(fn_list)
        fn = next(fn_list_copy)
        result = fn(*args_set)
        for fn in fn_list_copy:
            result = fn(result)
        yield result


def consume(iterator):
    """
    Consumes provided iterator
    """
    all(True for _ in iterator)



import time

from django import forms


class StartStopTime(forms.Form):

    max_range = None

    start_time = forms.IntegerField(required=False)
    stop_time = forms.IntegerField(required=False)

    def __init__(self, *args, **kwargs):
        self.pre_init()
        super().__init__(*args, **kwargs)

    def pre_init(self):
        self.declared_fields['stop_time'].required = \
            self.declared_fields['start_time'].required = \
            self.max_range is not None

    def clean(self):
        cleaned_data = super().clean()
        start_time = cleaned_data.get('start_time')
        stop_time = cleaned_data.get('stop_time')
        if None not in (stop_time, start_time):
            if start_time > stop_time:
                raise forms.ValidationError(
                    'start_time can\'t be greater then stop_time',
                )
            if self.max_range is not None and stop_time - start_time > self.max_range:
                raise forms.ValidationError(
                    'range between start_time and stop_time can\'t be greater '
                    'then {max_range}'.format(max_range=self.max_range),
                )
        return cleaned_data


class StartStopTimeDefaultStop(StartStopTime):

    def pre_init(self):
        super().pre_init()
        self.declared_fields['stop_time'].required = False

    def clean_stop_time(self):
        stop_time = self.cleaned_data['stop_time']
        return int(time.time()) if stop_time is None else stop_time



from django.contrib.postgres import lookups
from django.db import models


class PgTrigramCharField(models.CharField):
    pass


class PgTrigramTextField(models.TextField):
    pass


@PgTrigramCharField.register_lookup
@PgTrigramTextField.register_lookup
class SimilarTo(lookups.PostgresSimpleLookup):
    lookup_name = 'similar'
    operator = '%%'  # escaped '%'



import warnings

from django.contrib.postgres import fields
from django.db import models


class DateRangeField(fields.DateRangeField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'DateRangeField is deprecated and will be removed in '
            'rmr-django 2.0, use '
            'django.contrib.postgres.fields.DateRangeField instead',
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class DateTimeRangeField(fields.DateTimeRangeField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'DateTimeRangeField is deprecated and will be removed in '
            'rmr-django 2.0, use '
            'django.contrib.postgres.fields.DateTimeRangeField instead',
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class FloatRangeField(fields.FloatRangeField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'FloatRangeField is deprecated and will be removed in '
            'rmr-django 2.0, use '
            'django.contrib.postgres.fields.FloatRangeField instead',
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class BigIntegerRangeField(fields.BigIntegerRangeField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'BigIntegerRangeField is deprecated and will be removed in '
            'rmr-django 2.0, use '
            'django.contrib.postgres.fields.BigIntegerRangeField instead',
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class IntegerRangeField(fields.IntegerRangeField):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            'IntegerRangeField is deprecated and will be removed in '
            'rmr-django 2.0, use '
            'django.contrib.postgres.fields.IntegerRangeField instead',
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class Lower(models.Transform):
    lookup_name = 'lower'

    def as_sql(self, compiler, connection, *args, **kwargs):
        lhs, params = compiler.compile(self.lhs)
        return 'lower({lhs})'.format(lhs=lhs), params


class Upper(models.Transform):
    lookup_name = 'upper'

    def as_sql(self, compiler, connection, *args, **kwargs):
        lhs, params = compiler.compile(self.lhs)
        return 'upper({lhs})'.format(lhs=lhs), params


@DateRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.DateField()


@DateRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.DateField()


@DateTimeRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.DateTimeField()


@DateTimeRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.DateTimeField()


@FloatRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.FloatField()


@FloatRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.FloatField()


@BigIntegerRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.BigIntegerField()


@BigIntegerRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.BigIntegerField()


@IntegerRangeField.register_lookup
class DateRangeLowerTransform(Lower):
    output_field = models.IntegerField()


@IntegerRangeField.register_lookup
class DateRangeUpperTransform(Upper):
    output_field = models.IntegerField()



import warnings

from rmr.utils.db import BulkModelCreator

warnings.warn(
    'BulkModelCreator moved to rmr.utils.db, make sure you import it from '
    'the new place before upgrading to rmr-django 2.0',
    category=RuntimeWarning, stacklevel=2,
)



import warnings

from datetime import datetime, timedelta

import pytz

from django.utils import timezone, lru_cache

NOW = object()


def fromtimestamp(timestamp):
    if timestamp is None:
        return None
    return datetime.fromtimestamp(timestamp, tz=timezone.get_current_timezone())


def strptime(date_string, format, default_tz=timezone.utc):
    dt = datetime.strptime(date_string, format)
    if timezone.is_naive(dt):
        dt = timezone.make_aware(dt, timezone=default_tz)
    return timezone.localtime(dt)


@lru_cache.lru_cache(maxsize=None)
def get_timezones_with_hour(hour):
    """
    Returns list of timezones where current hour equals to provided value
    """

    now = timezone.now()

    return [
        tz for tz
        in pytz.common_timezones_set
        if timezone.localtime(now, pytz.timezone(tz)).hour == hour
    ]


def get_date_range(
    start_date: int = None,
    end_date: int = NOW,
    max_range: int = None,
):
    """
    Get date range

    :rtype: (datetime, datetime)
    """
    warnings.warn(
        'get_date_range is deprecated and will be removed in rmr-django 2.0, '
        'use rmr.forms.StartStopTime to validate request params',
        category=DeprecationWarning,
    )
    max_range = max_range or None
    if max_range and None in (start_date, end_date):
        raise ValueError(
            'Both start_date and end_date must be not None '
            'when max_range used'
        )

    if end_date is NOW:
        end = timezone.now()
    elif end_date is None:
        end = None
    else:
        end = fromtimestamp(end_date)

    if start_date is None:
        start = max_range and end - timedelta(seconds=max_range)
        return start, end
    else:
        start = fromtimestamp(start_date)

    if max_range and (end - start).seconds > max_range:
        raise ValueError('Date range exceeds max_range')

    return start, end



import contextlib
import functools
import logging
import warnings

from django.http import JsonResponse, HttpResponse, HttpRequest
from django.utils import six
from django.utils.decorators import classonlymethod
from django.views.decorators import http
from django.views.generic import View

from djangocache import cache_page

import rmr

from rmr.utils import decorators

request_logger = logging.getLogger('rmr.request')

response_logger = logging.getLogger('rmr.response')


class Json(View):

    http_code = 200

    logger = logging.getLogger('rmr.view')

    def __init__(self, request: HttpRequest=None, **kwargs):
        super().__init__(**kwargs)
        self.request = request

    @classmethod
    def _expires(cls, request, *args, **kwargs):
        if not six.get_function_defaults(cls.expires):
            warnings.warn(
                'In rmr-django 2.0 Json.expires() will be called with '
                'additional arguments: (request, *args, **kwargs)',
                category=RuntimeWarning,
            )
        return cls.expires()

    @classmethod
    def expires(cls, request=None, *args, **kwargs):
        """
        Lazy evaluated response cache TTL in seconds,
        corresponds to the Cache-Control's max-age value
        and also used to evaluate Expires header
        """
        return 0  # cache disabled

    @classmethod
    def last_modified(cls, request, *args, **kwargs):
        """
        Lazy evaluated value of Last-Modified header
        """
        pass

    @classmethod
    def etag(cls, request, *args, **kwargs):
        """
        Lazy evaluated value of ETag header
        """
        pass

    @classonlymethod
    def as_view(cls, **initkwargs):
        def normalize_key_prefix(key_prefix):
            def _key_prefix(request, *args, **kwargs):
                return str(
                    key_prefix(request, *args, **kwargs)
                ).replace(' ', '_')
            return _key_prefix

        patched_view = view = super().as_view(**initkwargs)

        patched_view = http.etag(cls.etag)(patched_view)
        patched_view = http.last_modified(cls.last_modified)(patched_view)
        patched_view = cache_page(
            cache_timeout=cls._expires,
            key_prefix=normalize_key_prefix(cls.last_modified),
        )(patched_view)

        view = decorators.replace_if(
            lambda request, *args, **kwargs: request.method in ('GET', 'HEAD'),
            replacement=patched_view,
        )(view)

        @functools.wraps(cls.as_view)
        def logging_view(request, *args, **kwargs):

            request_logger.debug(
                'request_method: %(request_method)s, '
                'request_path: %(request_path)s, '
                'request_headers: %(request_headers)s, '
                'request_params: %(request_params)s, '
                'request_data: %(request_data)s, ',
                dict(
                    request_method=request.method,
                    request_path=request.path,
                    request_headers=request.META,
                    request_params=request.GET,
                    request_data=request.POST,
                ),
            )

            response = view(request, *args, **kwargs)

            response_logger.debug(
                'response_code: %(response_code)s, '
                'response_headers: %(response_headers)s, '
                'response_data: %(response_data)s',
                dict(
                    response_code=response.status_code,
                    response_headers=response.items(),
                    response_data=response.content,
                ),
            )

            return response

        return logging_view

    def dispatch(self, request: HttpRequest, *args, **kwargs):
        self.request = request
        try:
            result = super().dispatch(request, *args, **kwargs)
            if isinstance(result, HttpResponse):
                return result
            http_code = self.http_code
            api_result = dict(
                data=result,
            )
        except rmr.Error as error:

            response_logger.log(
                error.level,
                '%(code)s: %(message)s',
                dict(message=error.message, code=error.code),
            )

            http_code = error.http_code
            api_result = dict(
                error=dict(
                    code=error.code,
                    description=error.message,
                ),
            )

        response = JsonResponse(api_result, status=http_code, safe=False)
        response.setdefault('Content-Length', len(response.content))
        return response

    @staticmethod
    def get_range(offset=None, limit=None, limit_default=None, limit_max=None):
        warnings.warn(
            'rmr.views.Json.get_range() is deprecated and will be removed '
            'in rmr-django 2.0, use rmr.utils.range.get_range() instead and/or '
            'rmr.forms.OffsetLimit form to validate offset/limit request '
            'params',
            category=DeprecationWarning,
        )
        start = 0
        stop = None
        with contextlib.suppress(ValueError):
            start = offset and int(offset) or start
            if start < 0:
                raise rmr.ClientError(
                    'Offset must be a positive number',
                    code='incorrect_offset',
                )
            limit = limit and int(limit) or limit_default
            if limit is not None:
                if limit <= 0:
                    raise rmr.ClientError(
                        'Limit must be a positive number',
                        code='incorrect_limit',
                    )
                if limit_max and limit > limit_max:
                    raise rmr.ClientError(
                        'Maximum value of limit must be '
                        'less then or equal {}'.format(limit_max),
                        code='max_limit_exceeded',
                    )
                stop = start + limit
            elif limit_max:
                raise rmr.ClientError(
                    'Limit must be provided',
                    code='limit_not_provided',
                )

            return start, stop

        raise rmr.ClientError(
            'Limit and offset must be a numbers if provided',
            code='incorrect_limit_or_offset',
        )



from distutils.core import setup
from setuptools import find_packages

with open('README.md') as description:
    long_description = description.read()

setup(
    name='rmr-django',
    version='1.1.5',
    author='Rinat Khabibiev',
    author_email='srenskiy@gmail.com',
    packages=list(map('rmr.'.__add__, find_packages(where='rmr'))) + ['rmr'],
    url='https://github.com/RedMadRobot/rmr_django',
    license='MIT',
    description='rmr_django',
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Environment :: Web Environment',
    ],
    install_requires=[
        'crcmod>=1.7,<2.0',
        'django-cache>=0.1,<0.2',
        'pytz',
    ],
)



import json

from django import http
from django.conf import settings

from rmr.types import JsonDict


class RequestDecoder:

    content_type = 'application/json'

    allowed_methods = {
        'POST', 'PUT', 'PATCH',
    }

    def process_request(self, request):
        if request.method not in self.allowed_methods:
            return
        content_type = request.META.get('CONTENT_TYPE', '')
        if not content_type.startswith(self.content_type):
            return
        encoding = request.encoding or settings.DEFAULT_CHARSET
        try:
            body = request.body.decode(encoding=encoding)
        except UnicodeDecodeError:
            return http.HttpResponseBadRequest('bad unicode')
        try:
            request.POST = self.json_decode(body)
        except ValueError:
            return http.HttpResponseBadRequest('malformed data')

    @staticmethod
    def json_decode(body):
        data = json.loads(body)
        if not isinstance(data, dict):
            # all data of type other then dict will be returned as is
            return data
        return JsonDict(data)



from django import http


class JsonDict(http.QueryDict):

    def __init__(self, json_dict=None, mutable=False):
        if json_dict:
            for key, value in json_dict.items():
                if isinstance(value, list):
                    # treat list values as set of values of MultiValueDict
                    self.setlist(key, value)
                else:
                    self[key] = value
        super().__init__(mutable=mutable)



import warnings

from rmr.middleware.json import RequestDecoder

warnings.warn(
    'RequestDecoder moved to rmr.middleware.json, make sure you import it from '
    'the new place before upgrading to rmr-django 2.0',
    category=RuntimeWarning,
)



def get_range(offset=None, limit=None, limit_default=None):
    """
    Converts offset/limit params to slice object

    For example, following two expressions are identical:
        some_list[get_range(offset=10, limit=5)]
        some_list[10:15]
    """
    start = offset or 0
    limit = limit or limit_default
    stop = limit and start + limit
    return slice(start, stop)



import functools

from django import forms, http

import rmr


def validate_request(*, get=forms.Form, post=forms.Form):
    def _decorator(view):
        @functools.wraps(view)
        def _view(request: http.HttpRequest, *args, **kwargs):
            form_get = get(request.GET)
            form_post = post(request.POST, request.FILES)
            if False in (form_get.is_valid(), form_post.is_valid()):
                raise rmr.ClientError(
                    code='validation_error',
                    message=dict(form_post.errors, **form_get.errors),
                )
            request.GET = form_get.cleaned_data
            request.POST = form_post.cleaned_data
            return view(request, *args, **kwargs)
        return _view
    return _decorator



from django.contrib.postgres import lookups
from django.db import models


class PgLtreeField(models.TextField):

    description = 'Ltree'

    def db_type(self, connection):
        return 'ltree'

    def get_internal_type(self):
        return "CharField"

    def formfield(self, **kwargs):
        return super(models.TextField, self).formfield()


@PgLtreeField.register_lookup
class AncestorOf(lookups.PostgresSimpleLookup):
    lookup_name = 'ancestor_of'
    operator = '@>'


@PgLtreeField.register_lookup
class DescendantOf(lookups.PostgresSimpleLookup):
    lookup_name = 'descendant_of'
    operator = '<@'


@PgLtreeField.register_lookup
class Match(lookups.PostgresSimpleLookup):
    lookup_name = 'match'
    operator = '~'
